%
%   Written by Michael Perrott for viewing CppSim/VppSim output
%   Distribution:  http://www.cppsim.com
%
%   This code is used as part of the CppSimView program
%   (cppsimview.m)
%
%   For those that use this code - please
%   keep my name attached to it (as currently done).  
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU Library General Public
% License along with this library; if not, write to the MxFree
% Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

% Eye diagrams for signal names in structure x.
% Usage: eyesig(x,period,time_offset,nodes)
% Where period is the data period, time_offset is the offset time, in seconds
% where you want to measure from, nodes is the expresion for the signals.
%  written by Mike Perrott
function status = eyesig(x,period_in,time_offset_in,nodes)

%%%%%%%%%%%%%%%%%%%%% Error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin ~= 4
     mssg = sprintf('Error:  must have 4 arguments: eyesig(x,period,time_offset,''nodes'')');
     error_box(mssg,'Error using ''eyesig''');
     return 
end

if ischar(period_in) == 1
   period = str2num(period_in);
else
   period = period_in;
end
if ischar(time_offset_in) == 1
   time_offset = str2num(time_offset_in);
else
   time_offset = time_offset_in;
end

t_in = x(1).data;
if length(t_in) == 0
    return
end

ind = find(t_in >= time_offset);

if length(ind) > 0
   start_i = ind(1);
else
   err_message = sprintf('Error: starting time offset of ''%s'' is too large --- max time in simulation = %5.3e',time_offset_in,max(t_in));     
   error_box(err_message,'Error in Plot Expression');
   status = 0;
   return;
end

%%%%%%%%%%%%%%%%%%%%%% error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%
 if length(t_in) < 2
     mssg = sprintf('Error:  must have at least 2 time samples');
     error_box(mssg,'Error using ''eyesig''');
     return;
 end
Ts = (t_in(length(t_in))-t_in(1))/(length(t_in)-1);

period_samples = round(period/Ts);

if period_samples <= 4
   err_message = sprintf('Error: resulting number of samples per period = %d --- it must be > 4',period_samples);     
   error_box(err_message,'Error in Plot Expression');
   status = 0;
   return;
end

%%%%%%%%%%%%%%%%%%%%% Labels  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

x_labels = {'Time (seconds)' 'Time (milliseconds)' 'Time (microseconds)' ...
            'Time (nanoseconds)' 'Time (picoseconds)'};
x_limits = [1 1e-3 1e-6 1e-9 1e-12];

ylabel_name = ' ';
if sum(nodes == ';') > 0 | sum(nodes == ',') > 0
   title_name = 'CppSim Simulated Eye Diagrams';
else
   title_name = 'CppSim Simulated Eye Diagram';
end

clf
status = -1;

%%%%%%%%%%%%%%%%%%%%% x definition %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
t_in = form_x_eye_vector(t_in,period,time_offset);
[xscaled,xlabel_name] = scale_and_label_x(t_in,x_limits,x_labels);

%%%%%%%%%%%%%%%%%%%%% Loop through subplots   %%%%%%%%%%%%%%%%%%%%%%
nsp = 1 + sum(nodes == ';');	% number of subplots
subplot(nsp,1,1);

for i = 1:nsp
    subplot(nsp,1,i);
%%%%%%%%%%%%%%%%%%%%% y definition %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [ysig,legend_string] = create_ysignals_and_legend(x,nodes,i);
    if length(ysig) == 0
       return
    end
    ysig = form_y_eye_vectors(x(1).data,ysig,period,time_offset);
%%%%%%%%%%%%%%%%%%%%% Plot function %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    deci_plot(xscaled,ysig,i);

%%%%%%%%%%%%%%%%%%%%% Labeling %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    h_leg = legend(legend_string);
    set(h_leg,'TextColor',[1 1 1]);
end

create_title(title_name);
create_ylabel(ylabel_name);
create_xlabel(xlabel_name); 

status = 0;
