%
%   Written by Michael Perrott for viewing CppSim/VppSim output
%   Distribution:  http://www.cppsim.com
%
%   This code is used as part of the CppSimView program
%   (cppsimview.m)
%
%   For those that use this code - please
%   keep my name attached to it (as currently done).  
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU Library General Public
% License along with this library; if not, write to the MxFree
% Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

% Eye diagrams for signal names in structure x.
% Usage: eyesig(x,period,time_offset,nodes)
% Where period is the data period, time_offset is the offset time, in seconds
% where you want to measure from, nodes is the expresion for the signals.
%  written by Mike Perrott
function status = eyesig_gmsk(x,period_in,time_offset_in,nodes)

%%%%%%%%%%%%%%%%%%%%% Error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if nargin ~= 4
     mssg = sprintf('Error:  must have 4 arguments: eyesig_gmsk(x,period,time_offset,''nodes'')');
     error_box(mssg,'Error using ''eyesig_gmsk''');
     return 
end

nsp = 1 + sum(nodes == ';');	% number of subplots
if nsp ~= 2
     mssg = sprintf('Error:  must include BOTH I and Q signals: eyesig_gmsk(x,period,time_offset,''node_I;node_Q'')');
     error_box(mssg,'Error using ''eyesig_gmsk''');
     return 
end

nca = sum(nodes == ',');	% number of commas
if nca ~= 0
     mssg = sprintf('Error:  no commas are allowed in ''nodes'' expression: eyesig_gmsk(x,period,time_offset,''nodes'')');
     error_box(mssg,'Error using ''eyesig_gmsk''');
     return 
end


if ischar(period_in) == 1
   period = str2num(period_in);
else
   period = period_in;
end
if ischar(time_offset_in) == 1
   time_offset = str2num(time_offset_in);
else
   time_offset = time_offset_in;
end

t_in = x(1).data;
if length(t_in) == 0
    return
end

ind = find(t_in >= time_offset);

if length(ind) > 0
   start_i = ind(1);
else
   err_message = sprintf('Error: starting time offset of ''%s'' is too large --- max time in simulation = %5.3e',time_offset_in,max(t_in));     
   error_box(err_message,'Error in Plot Expression');
   status = 0;
   return;
end

%%%%%%%%%%%%%%%%%%%%%% error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%
 if length(t_in) < 2
     mssg = sprintf('Error:  must have at least 2 time samples');
     error_box(mssg,'Error using ''eyesig''');
     return;
 end
Ts = (t_in(length(t_in))-t_in(1))/(length(t_in)-1);

period_samples = round(period/Ts);

if period_samples <= 4
   err_message = sprintf('Error: resulting number of samples per period = %d --- it must be > 4',period_samples);     
   error_box(err_message,'Error in Plot Expression');
   status = 0;
   return;
end

%%%%%%%%%%%%%%%%%%%%% Labels  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

x_labels = {'Time (seconds)' 'Time (milliseconds)' 'Time (microseconds)' ...
            'Time (nanoseconds)' 'Time (picoseconds)'};
x_limits = [1 1e-3 1e-6 1e-9 1e-12];

ylabel_name = ' ';
if sum(nodes == ';') > 0 | sum(nodes == ',') > 0
   title_name = 'CppSim Simulated Eye Diagrams';
else
   title_name = 'CppSim Simulated Eye Diagram';
end

clf
status = -1;

%%%%%%%%%%%%%%%%%%%%% x definition %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
t_in = form_x_eye_vector(t_in,period,time_offset);
[xscaled,xlabel_name] = scale_and_label_x(t_in,x_limits,x_labels);

%%%%%%%%%%%%%%%%%%%%% Obtain subplot I/Q signals   %%%%%%%%%%%%%%%%%%%%%%
[ysigi,legend_stringi] = create_ysignals_and_legend(x,nodes,1);
if length(ysigi) == 0
    return
end
[ysigq,legend_stringq] = create_ysignals_and_legend(x,nodes,2);
if length(ysigq) == 0
    return
end

%%%%%%%%%%%%%%%%%%%% Rotate for best eye diagram  %%%%%%%%%%%%%%%%%%%%
[ysigi_rot,ysigq_rot,theta] = gmsk(ysigi,ysigq,period/Ts);

%%%%%%%%%%%%%%%%%%%%% Plot eye diagrams %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
ysig = form_y_eye_vectors(x(1).data,ysigi_rot,period,time_offset);
subplot(nsp,1,1);
deci_plot(xscaled,ysig,1);
legend_string = cellstr(sprintf('%s: %5.1f degree shift',char(legend_stringi),theta*180/pi));
legend(legend_string);

ysig = form_y_eye_vectors(x(1).data,ysigq_rot,period,time_offset);
subplot(nsp,1,2);
deci_plot(xscaled,ysig,2);
legend_string = cellstr(sprintf('%s: %5.1f degree shift',char(legend_stringq),theta*180/pi));
legend(legend_string);
%%%%%%%%%%%%%%%%%%%%% Labeling %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%




create_title(title_name);
create_ylabel(ylabel_name);
create_xlabel(xlabel_name); 

status = 0;


function [ysigi_rot,ysigq_rot,theta] = gmsk(ysigi_in,ysigq_in,period)

mid = (max(ysigi_in) + min(ysigi_in))/2;
ysigi = ysigi_in - mid;
mid = (max(ysigq_in) + min(ysigq_in))/2;
ysigq = ysigq_in - mid;


theta_vec = (0:0.1:1.0)*pi/2;
min_i = find_best_theta(theta_vec,ysigi,ysigq,period);

theta_min = theta_vec(min_i)-0.1*pi/2;
theta_max = theta_vec(min_i)+0.1*pi/2;
delta_theta = (theta_max - theta_min)/20;

theta_vec = theta_min:delta_theta:theta_max;
min_i = find_best_theta(theta_vec,ysigi,ysigq,period);

theta = theta_vec(min_i);
ysigi_rot = ysigi*cos(theta) - ysigq*sin(theta);
ysigq_rot = ysigi*sin(theta) + ysigq*cos(theta);


function [min_i] = find_best_theta(theta_vec,ysigi,ysigq,period_in) 

period = period_in/2;

min_i = 1;
min_var = 10e6;
for i = 1:length(theta_vec)
   theta = theta_vec(i);
   ysigi_rot = ysigi*cos(theta) - ysigq*sin(theta);
   ysigq_rot = ysigi*sin(theta) + ysigq*cos(theta);

   edge_veci = zeros(length(ysigi),1);
   edge_vecq = zeros(length(ysigi),1);
   ki = 1;
   kq = 1;
   count = 1;
   for j = 2:length(ysigi)
     if sign(ysigi_rot(j)) ~= sign(ysigi_rot(j-1))
        edge_time = abs(ysigi_rot(j))/(abs(ysigi_rot(j))+abs(ysigi_rot(j-1)));
        edge_veci(ki) = count-edge_time;
        if (edge_veci(ki) <= 0)
           edge_veci(ki) = edge_veci(ki) + period;
        end
        ki = ki + 1;
     end
     if sign(ysigq_rot(j)) ~= sign(ysigq_rot(j-1))
        edge_time = abs(ysigq_rot(j))/(abs(ysigq_rot(j))+abs(ysigq_rot(j-1)));
        edge_vecq(kq) = count-edge_time;
        if (edge_vecq(kq) <= 0)
           edge_vecq(kq) = edge_vecq(kq) + period;
        end
        kq = kq + 1;
     end
     if (count >= period)
        count = count-period;
     end
     count = count + 1;
   end

   edge_veci_n = edge_veci(2:ki-1);
   edge_vecq_n = edge_vecq(2:kq-1);

   sig = var(edge_veci_n) + var(edge_vecq_n);
   if sig < min_var
        min_var = sig;
        min_i = i;
   end
end

