%
%   Written by Michael Perrott for viewing CppSim/VppSim output
%   Distribution:  http://www.cppsim.com
%
%   This code is used as part of the CppSimView program
%   (cppsimview.m)
%
%   For those that use this code - please
%   keep my name attached to it (as currently done).  
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU Library General Public
% License along with this library; if not, write to the MxFree
% Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

%%% This Matlab code is used to plot modulated spectrum of a synthesizer
%  written by Michael Perrott for CppSimView in March 2004
function status = plot_pll_mod_spectrum(x,f_span_in,nodes)

%%%%%%%%%%%%%%%%%%%%% Error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 if nargin ~= 3
     mssg = sprintf('Error:  must have 3 arguments: plot_pll_mod_spectrum(x,f_span,''nodes'')');
     error_box(mssg,'Error using ''plot_pll_mod_spectrum''');
     return 
 end

 clf

 if ischar(f_span_in) == 1
    f_span = str2num(f_span_in);
 else
    f_span = f_span_in;
 end

%%%%%%%%%%%%%%%%%%%%% Labels  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


x_labels = {'Frequency Offset from Carrier (GHz)' ...
            'Frequency Offset from Carrier (MHz)' ...
            'Frequency Offset from Carrier (kHz)' ...
            'Frequency Offset from Carrier (Hz)'};
x_limits = [1e9 1e6 1e3 1];

ylabel_name = 'Spectrum (dBc/Hz)';
title_name = 'CppSim Simulated Modulated PLL Spectrum';

clf
status = -1;

%%%%%%%%%%%%%%%%%%%%% x definition (first part) %%%%%%%%%%%%%%%%%%%%%%%%%%%%

t = x(1).data;
if length(t) < 20
    return
end

%  set parameters
Ts_vec = filter([1 -1],1,t);
if length(Ts_vec) > 10010
   Ts = mean(Ts_vec(10:10010));
else
   Ts = mean(Ts_vec(10:length(Ts_vec)));
end

Kv = 1.0; % VCO gain (Hz/V) always assumed to be 1.0


%%%%%%%%%%%%%%%%%%%%% Loop through subplots   %%%%%%%%%%%%%%%%%%%%%%
nsp = 1 + sum(nodes == ';');	% number of subplots
subplot(nsp,1,1);

for i = 1:nsp
    subplot(nsp,1,i);
%%%%%%%%%%%%%%%%%%%%% y definition %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [ysig,legend_string] = create_ysignals_and_legend(x,nodes,i);
    if length(ysig) == 0
       return
    end

%   create VCO output phase based on integration of vin
    for k = 1:size(ysig,2)

       phase = filter(Ts*2*pi*Kv,[1 -1],ysig(:,k)-mean(ysig(:,k)));

       baseband_out = cos(phase)+j*sin(phase);
       baseband_out = baseband_out - mean(baseband_out);
%  calculate output spectrum referred to baseband
       [Pxx,f] = psd(sqrt(Ts)*baseband_out,2^16,1/Ts,2^16,'mean');
       Pxx = fftshift(Pxx);
       f = f - 1.0/(2.0*Ts);

%%%%%%%%%%%%%%%%%%%%%% error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%
       if i == 1 & k == 1
          ind = find(f >= f_span);
          if length(ind) > 0
              i_high = ind(1);
          else
              mssg = sprintf('Error:  f_span = ''%s'' is too high --- max frequency is only %5.3e',f_span_in,max(f));
              error_box(mssg,'Error using ''plot_pll_mod_spectrum''');
              return;
          end 

          ind = find(f <= -f_span);
          if length(ind) > 0
              i_low = ind(length(ind));
          else
              mssg = sprintf('Error:  f_span = ''%s'' is too high --- max frequency is only %5.3e',f_span_in,max(f));
              error_box(mssg,'Error using ''plot_pll_mod_spectrum''');
              return;
          end 
       end
       if k == 1
          spec_dens = zeros(length(Pxx),size(ysig,2));
       end

       spec_dens(:,k) = Pxx;
    end
%%%%%%%%%%%%%%%%%%%%%% extraction and scaling  %%%%%%%%%%%%%%%%%%%%%%%%%%%
    spec_dens = 10*log10(spec_dens(i_low:i_high,:));
    if i == 1
        f = f(i_low:i_high);
        [f_new,xlabel_name] = scale_and_label_x(f,x_limits,x_labels);
    end
%%%%%%%%%%%%%%%%%%%%% Plot function %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    deci_plot(f_new,spec_dens,i);

%%%%%%%%%%%%%%%%%%%%% Labeling %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    h_leg = legend(legend_string);
    set(h_leg,'TextColor',[1 1 1]);
end

create_title(title_name);
create_ylabel(ylabel_name);
create_xlabel(xlabel_name); 

status = 0;
