%
%   Written by Michael Perrott for viewing CppSim/VppSim output
%   Distribution:  http://www.cppsim.com
%
%   This code is used as part of the CppSimView program
%   (cppsimview.m)
%
%   For those that use this code - please
%   keep my name attached to it (as currently done).  
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU Library General Public
% License along with this library; if not, write to the MxFree
% Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

%%% This Matlab code is used to plot phase noise of a synthesizer
%  written by Michael Perrott for CppSimView in March 2004
function status = plot_pll_phasenoise(x,f_low_in,f_high_in,nodes)

global right_axis_offset_global

%%%%%%%%%%%%%%%%%%%%% Error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 if nargin ~= 4
     mssg = sprintf('Error:  must have 4 arguments: plot_pll_phase_noise(x,f_low,f_high,''nodes'')');
     error_box(mssg,'Error using ''plot_pll_phase_noise''');
     return 
 end

 if exist('psd') ~= 2 && exist('pwelch') ~=2
     mssg = sprintf('Error:  you must have the Matlab Signal Processing Toolbox to use plot_pll_phase_noise()');
     error_box(mssg,'Error using ''plot_pll_phase_noise''');
     return 
 end

 clf

 if ischar(f_low_in) == 1
    f_low = str2num(f_low_in);
 else
    f_low = f_low_in;
 end
 if ischar(f_high_in) == 1
    f_high = str2num(f_high_in);
 else
    f_high = f_high_in;
 end
%%%%%%%%%%%%%%%%%%%%% Labels  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


x_labels = {'Frequency Offset from Carrier (GHz)' ...
            'Frequency Offset from Carrier (MHz)' ...
            'Frequency Offset from Carrier (kHz)' ...
            'Frequency Offset from Carrier (Hz)'};
x_limits = [1e9 1e6 1e3 1];

ylabel_name = 'L(f) (dBc/Hz)';
title_name = 'CppSim Simulated Phase Noise';

clf
status = -1;

%%%%%%%%%%%%%%%%%%%%% x definition (first part) %%%%%%%%%%%%%%%%%%%%%%%%%%%%

t = x(1).data;
if length(t) < 20
    return
end

%  set parameters
Ts_vec = filter([1 -1],1,t);
if length(Ts_vec) > 10010
   Ts = mean(Ts_vec(10:10010));
else
   Ts = mean(Ts_vec(10:length(Ts_vec)));
end

Kv = 1.0; % VCO gain (Hz/V) always assumed to be 1.0


%%%%%%%%%%%%%%%%%%%%% Loop through subplots   %%%%%%%%%%%%%%%%%%%%%%
nsp = 1 + sum(nodes == ';');	% number of subplots
subplot(nsp,1,1);

for i = 1:nsp
    subplot(nsp,1,i);
%%%%%%%%%%%%%%%%%%%%% y definition %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [ysig,legend_string] = create_ysignals_and_legend(x,nodes,i);
    if length(ysig) == 0
       return
    end

%   create VCO output phase based on integration of vin
    for k = 1:size(ysig,2)
       phase = filter(Ts*2*pi*Kv,[1 -1],ysig(:,k)-mean(ysig(:,k)));
       if i == 1 & k == 1
       %%% take psd of reference signal to determine how to scale spurs
           ref_sig = sqrt(Ts)*2*cos(2*pi/10*(1:length(phase)));
           if exist('psd') == 2
              [ref_spec,f] = psd(ref_sig,2^16,1/Ts,2^16,'mean');
           else
	     ref_sig = dtrend(ref_sig);
             [ref_spec,f] = pwelch(ref_sig,2^16,[],2^16,1/Ts);
           end
           right_axis_offset_global = 10*log10(1/max(ref_spec));
           spur_offset = round(right_axis_offset_global);
       end
%      calculate L(f)
       if exist('psd') == 2
           [Pxx,f] = psd(sqrt(Ts)*phase,2^16,1/Ts,2^16,'mean');
       else
	  phase = dtrend(phase);
          [Pxx,f] = pwelch(phase,2^16,[],2^16,1/Ts);
       end


%%%%%%%%%%%%%%%%%%%%%% error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%
       if i == 1 & k == 1
          ind = find(f >= f_low);
          if length(ind) > 0
              i_low = ind(1);
          else
              mssg = sprintf('Error:  f_low = ''%s'' is too high --- max frequency is only %5.3e',f_low_in,max(f));
              error_box(mssg,'Error using ''plot_pll_phase_noise''');
              return;
          end 

          ind = find(f >= f_high);
          if length(ind) > 0
             i_high = ind(1);
          else
             mssg = sprintf('Error:  f_high = ''%s'' is too high --- max frequency is only %5.3e',f_high_in,max(f));
             error_box(mssg,'Error using ''plot_pll_phase_noise''');
             return
          end

          if i_high <= i_low
             mssg = sprintf('Error:  f_high = ''%s'' must be higher than f_low = ''%s''',f_high_in,f_low_in);
             error_box(mssg,'Error using ''plot_pll_phase_noise''');
             return
          end
       end
       if k == 1
          spec_dens = zeros(length(Pxx),size(ysig,2));
       end

       spec_dens(:,k) = Pxx;
    end
%%%%%%%%%%%%%%%%%%%%%% extraction and scaling  %%%%%%%%%%%%%%%%%%%%%%%%%%%
    spec_dens = 10*log10(spec_dens(i_low:i_high,:));
    if i == 1
        f = f(i_low:i_high);
        [f_new,xlabel_name] = scale_and_label_x(f,x_limits,x_labels);
    end
%%%%%%%%%%%%%%%%%%%%% Plot function %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    deci_plot(f_new,spec_dens,i,'logx,right_ax',spur_offset,'Spurs (dBc)');

%%%%%%%%%%%%%%%%%%%%% Labeling %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    h_leg = legend(legend_string);
    set(h_leg,'TextColor',[1 1 1]);
end

create_title(title_name);
create_ylabel(ylabel_name);
create_xlabel(xlabel_name); 

status = 0;
