%
%   Written by Michael Perrott for viewing CppSim/VppSim output
%   Distribution:  http://www.cppsim.com
%
%   This code is used as part of the CppSimView program
%   (cppsimview.m)
%
%   For those that use this code - please
%   keep my name attached to it (as currently done).  
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU Library General Public
% License along with this library; if not, write to the MxFree
% Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

%%% This Matlab code is used to plot power spectral density
% Usage: plot_psd_sqrt(x, nodes) or plot_psd_sqrt(x, nodes, plot_args)
%        where plot_args is 'logx', 'logy', or 'logxy'.
% In nodes, commas delimit curves to be overlayed and semicolons delimit
% seperate subplots on the same figure.  E.g., plot_psd_sqrt(x,'v1,v2;v3')
% creates two subplots, with voltages v1 and v2 overlayed.  Also note
% that numeric node names should be prepended by '@' to distinguish
% them from constants.
%  written by Mike Perrott


function status = plot_psd_sqrt(x,nodes,plot_args)

global right_axis_offset_global

%%%%%%%%%%%%%%%%%%%%% Error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if nargin < 2 | nargin > 3
     mssg = sprintf('Error:  must have either 2 or 3 arguments: plot_psd_sqrt(x,''nodes'',[plot_args])');
     error_box(mssg,'Error using ''plot_psd_sqrt''');
     return 
end

if exist('psd') ~= 2 && exist('pwelch') ~=2
     mssg = sprintf('Error:  you must have the Matlab Signal Processing Toolbox to use plot_psd_sqrt()');
     error_box(mssg,'Error using ''plot_psd_sqrt''');
     return 
 end

clf

%%%%%%%%%%%%%%%%%%%%% Labels  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


x_labels = {'Frequency (GHz)' ...
            'Frequency (MHz)' ...
            'Frequency (kHz)' ...
            'Frequency (Hz)'};
x_limits = [1e9 1e6 1e3 1];

ylabel_name = 'V/sqrt(Hz)';
title_name = 'One-Sided PSD';

clf
status = -1;

%%%%%%%%%%%%%%%%%%%%% x definition (first part) %%%%%%%%%%%%%%%%%%%%%%%%%%%%

t = x(1).data;
if length(t) < 20
    return
end

%  set parameters
Ts_vec = filter([1 -1],1,t);
if length(Ts_vec) > 10010
   Ts = mean(Ts_vec(10:10010));
else
   Ts = mean(Ts_vec(10:length(Ts_vec)));
end

%%%%%%%%%%%%%%%%%%%%% Loop through subplots   %%%%%%%%%%%%%%%%%%%%%%
nsp = 1 + sum(nodes == ';');	% number of subplots
subplot(nsp,1,1);

for i = 1:nsp
    subplot(nsp,1,i);
%%%%%%%%%%%%%%%%%%%%% y definition %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [ysig,legend_string] = create_ysignals_and_legend(x,nodes,i);
    if length(ysig) == 0
       return
    end

%   calculate power spectral density
    for k = 1:size(ysig,2)
       %%% calculate psd
       if exist('psd') == 2
          [Pxx,f] = psd(sqrt(Ts)*(ysig(:,k)-mean(ysig(:,k))),2^16,1/Ts,2^16,'mean');
       else
	  ysig_dtrend = dtrend(ysig(:,k)-mean(ysig(:,k)));
          [Pxx,f] = pwelch(ysig_dtrend,2^16,[],2^16,1/Ts);
       end

       if k == 1
          spec_dens = zeros(length(Pxx)-1,size(ysig,2));
       end

       spec_dens(:,k) = sqrt(2*abs(Pxx(2:end)));
    end
%%%%%%%%%%%%%%%%%%%%%% extraction and scaling  %%%%%%%%%%%%%%%%%%%%%%%%%%%
    if i == 1
        [f_new,xlabel_name] = scale_and_label_x(f(2:end),x_limits,x_labels);
    end
%%%%%%%%%%%%%%%%%%%%% Plot function %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if nargin > 2
       plot_args = strcat(plot_args,',nodec');
       deci_plot(f_new,spec_dens,i,plot_args);
    else
       deci_plot(f_new,spec_dens,i,'nodec');
    end
    h_leg = legend(legend_string);
    set(h_leg,'TextColor',[1 1 1]);

%%%%%%%%%%%%%%%%%%%%% Labeling %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    legend(legend_string);
end

create_title(title_name);
create_ylabel(ylabel_name);
create_xlabel(xlabel_name); 

status = 0;
