%
%   Written by Michael Perrott for viewing CppSim/VppSim output
%   Distribution:  http://www.cppsim.com
%
%   This code is used as part of the CppSimView program
%   (cppsimview.m)
%
%   For those that use this code - please
%   keep my name attached to it (as currently done).  
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU Library General Public
% License along with this library; if not, write to the MxFree
% Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%

% Plot expressions invoking signal names from structure x.
% Usage: plotsig(x, nodes) or plotsig(x, nodes, plot_args)
%        where plot_args is 'logx', 'logy', or 'logxy'.
% In nodes, commas delimit curves to be overlayed and semicolons delimit
% seperate subplots on the same figure.  E.g., plotsig(x,'v1,v2;v3')
% creates two subplots, with voltages v1 and v2 overlayed.  Also note
% that numeric node names should be prepended by '@' to distinguish
% them from constants.
%  written by Mike Perrott

function status = plotsig(x,nodes,plot_args)

%%%%%%%%%%%%%%%%%%%%% Error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if nargin < 2 | nargin > 3
     mssg = sprintf('Error:  must have either 2 or 3 arguments: plotsig(x,''nodes'',[plot_args])');
     error_box(mssg,'Error using ''plotsig''');
     return 
end
%%%%%%%%%%%%%%%%%%%%% Labels  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

x_labels = {'Time (seconds)' 'Time (milliseconds)' 'Time (microseconds)' ...
            'Time (nanoseconds)' 'Time (picoseconds)'};
x_limits = [1 1e-3 1e-6 1e-9 1e-12];

ylabel_name = ' ';
if sum(nodes == ';') > 0 | sum(nodes == ',') > 0
   title_name = 'CppSim Simulated Signals';
else
   title_name = 'CppSim Simulated Signal';
end

clf
status = -1;

%%%%%%%%%%%%%%%%%%%%% x definition %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
xsig = x(1).data;
%%%%%%%%%%%%%%%%%%%%% Error checking %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if length(xsig) < 2
     mssg = sprintf('Error:  must have at least 2 time samples');
     error_box(mssg,'Error using ''plotsig''');
     return 
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
Ts = (xsig(length(xsig))-xsig(1))/(length(xsig)-1);

start_val = xsig(1);
final_val = xsig(length(xsig));
xsig_temp = Ts*(0:(length(xsig)-1))+start_val;
xsig_temp = xsig_temp';
%figure(2)
%plot(1:length(xsig),abs(xsig-xsig_temp),1:length(xsig),1e-6*xsig_temp)
%figure(1)
if min(1e-6*xsig_temp - abs(xsig-xsig_temp)) >= 0.0
   xsig = xsig_temp;
%   disp('replaced time vector');
end

[xscaled,xlabel_name] = scale_and_label_x(xsig,x_limits,x_labels);

%%%%%%%%%%%%%%%%%%%%% Loop through subplots   %%%%%%%%%%%%%%%%%%%%%%
nsp = 1 + sum(nodes == ';');	% number of subplots
subplot(nsp,1,1);

for i = 1:nsp
    subplot(nsp,1,i);
%%%%%%%%%%%%%%%%%%%%% y definition %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    [ysig,legend_string] = create_ysignals_and_legend(x,nodes,i);
    if length(ysig) == 0
       return
    end

%%%%%%%%%%%%%%%%%%%%% Plot function %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    if nargin > 2
	deci_plot(xscaled,ysig,i,plot_args);
    else
        deci_plot(xscaled,ysig,i);
    end

%%%%%%%%%%%%%%%%%%%%% Labeling %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

    h_leg = legend(legend_string);
    set(h_leg,'TextColor',[1 1 1]);
end

create_title(title_name);
create_ylabel(ylabel_name);
create_xlabel(xlabel_name); 

status = 0;
